package TestUtils;

use strict;
use vars qw($VERSION
	    @ISA
	    @EXPORT
	    @EXPORT_OK
	    $DOM
	    $PERSONAL
	   );
use Carp;
require Exporter;

@ISA = qw(Exporter);

@EXPORT_OK = qw(result warning error fatal_error $DOM $PERSONAL);

BEGIN {
  $DOM = new XML::Xerces::DOMParser;
  $DOM->setToCreateXMLDeclTypeNode(1);

  my $error_handler = new XML::Xerces::perlErrorHandler;
  $error_handler->setWarningFunction(\&warning);
  $error_handler->setErrorFunction(\&error);
  $error_handler->setFatalErrorFunction(\&fatal_error);
  $DOM->setErrorHandler($error_handler);

  open(PERSONAL, "personal.xml")
    or die "Couldn't open personal.xml for reading";
  $/ = undef;
  $PERSONAL = <PERSONAL>;
  close PERSONAL;
}

sub warning {
  print STDERR "WARNING:\n";
  print STDERR "LINE:    ", $_[0]->getLineNumber, "\n";
  print STDERR "COLUMN:  ", $_[0]->getColumnNumber, "\n";
  print STDERR "MESSAGE: ", $_[0]->getMessage, "\n";
  exit 1;
}

sub error {
  print STDERR "ERROR:\n";
  print STDERR "LINE:    ", $_[0]->getLineNumber, "\n";
  print STDERR "COLUMN:  ", $_[0]->getColumnNumber, "\n";
  print STDERR "MESSAGE: ", $_[0]->getMessage, "\n";
  exit 1;
}

sub fatal_error {
  print STDERR "FATAL ERROR:\n";
  print STDERR "LINE:    ", $_[0]->getLineNumber, "\n";
  print STDERR "COLUMN:  ", $_[0]->getColumnNumber, "\n";
  print STDERR "MESSAGE: ", $_[0]->getMessage, "\n";
  exit 1;
}

sub result {
  my ($cond,$i) = @_;
  print STDOUT "not " unless $cond;
  print STDOUT "ok ", $i, "\n";
}
